import numpy as np


class FeatureGenerator:
    """
    FG do NOT handle joining data
    """

    def __init__(self,
                 past_fields,    # = ['open', 'close', 'high', 'low', 'amount'],
                 past_dts,       # = [-i for i in range(10)][::-1],
                 future_dts,     # = [1, 2, 3, 5, 10],
                 future_fields,  # = ['open', 'close', 'high', 'low']
                ):
        self.past_fields = past_fields
        self.future_fields = future_fields
        self.past_dts = past_dts
        self.future_dts = future_dts

        self.fea_tuple = None


    def make_label_plan(self):
        future_dts = self.future_dts
        fields = self.future_fields
        label_tuple = []
        for dt in future_dts:
            for field in fields:
                base_name = f'{field}_0'
                val_name = f'{field}_{dt}'
                incr_name = f'{field}_{dt}_incr'
                label_tuple.append((base_name, val_name, incr_name))
        return label_tuple

    def make_fea_plan(self):
        past_dts = self.past_dts
        fields = self.past_fields
        fea_tuple = []
        for last_dt, dt in zip(past_dts[:-1], past_dts[1:]):
            for field in fields:
                base_name = f'{field}_{last_dt}'
                val_name = f'{field}_{dt}'
                incr_name = f'd_{field}_{dt}'
                fea_tuple.append((base_name, val_name, incr_name))
        return fea_tuple

    def generate_feature(self, df):
        res = df[['date', 'code']].copy()
        label_tuple = self.make_label_plan()
        fea_tuple = self.make_fea_plan()
        for base_name, val_name, incr_name in label_tuple + fea_tuple:
            res[incr_name] = df[val_name] / df[base_name] - 1
        return res



from chinese_calendar import is_workday
from datetime import datetime
import pandas as pd


def is_trade_day(date):
    # 检查日期是否是工作日且不是周末
    if is_workday(date) and date.weekday() < 5:  # 周一到周五
        return True
    return False


def shift_join(df, fields, shift, shift_by='date', prefix=None):
    """
    :param df: a pandas data frame
    :param fields: a list
    :param shift: int
    :param shift_by: a column in df
    :param prefix: prefix for names of shifted fields
    :return: df with shifted field
    """
    df = df.sort_values(by=shift_by, ascending=True)
    df['index'] = list(range(df.shape[0]))
    dff = df[fields + ['index']].copy()
    dff['index'] = dff['index'] - shift

    def _rename(field):
        if prefix is None:
            return f'{field}_{shift}'
        else:
            return f'{prefix}_{field}_{shift}'

    rename_dict = {field: _rename(field) for field in fields}
    dff = dff.rename(columns=rename_dict)
    df = df.merge(dff, how='left', on='index')
    return df





import numpy as np


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def logit(x):
    return np.log(x / (1-x))


def get_auc(pred, label):
    assert len(pred) == len(label), f'len(pred) = {len(pred)}, len(label) = {len(label)}'
    tp = [x for x, y in zip(pred, label) if y == 1]     # true positive
    fp = [x for x, y in zip(pred, label) if y == 0]     # false positive
    tp = sorted(tp)
    fp = sorted(fp)
    ordered_pair = 0
    neg_idx = 0
    for i, x in enumerate(tp):
        while neg_idx < len(fp) and fp[neg_idx] < x:
            neg_idx += 1
        ordered_pair = neg_idx * i
    return ordered_pair / (len(fp) * len(fp))







from functools import reduce


class Stream:
    def __init__(self, obj):
        if type(obj) == int:
            self.iterable = range(obj)
        else:
            self.iterable = obj

    # transform
    def map(self, fn):
        self.iterable = map(fn, self.iterable)
        return self

    def filter(self, fn):
        self.iterable = filter(fn, self.iterable)
        return self

    def unique(self):
        self.iterable = list(set(self.iterable))
        return self

    # group
    def collect_list(self, key_fn=None):
        if key_fn is not None:
            self.iterable = map(lambda row: (key_fn(row), row), self.iterable)

        collect_dict = dict()
        for k, v in self.iterable:
            if k in collect_dict:
                collect_dict[k].append(v)
            else:
                collect_dict[k] = [v]
        self.iterable = collect_dict.items()
        return self

    def agg(self, agg_fn):
        res_dict = dict()
        for k, v_list in self.iterable:
            try:
                res_dict[k] = reduce(agg_fn, v_list)
            except:
                print(k, v_list)
        self.iterable = res_dict.items()
        return self

    def keys(self):
        self.iterable = map(lambda kv: kv[0], self.iterable)
        return self

    def values(self):
        self.iterable = map(lambda kv: kv[1], self.iterable)
        return self

    # execute
    def len(self):
        return len(list(self.iterable))

    def sum(self):
        return sum(list(self.iterable))

    def max(self):
        return max(list(self.iterable))

    def min(self):
        return min(list(self.iterable))

    def tolist(self):
        return list(self.iterable)

    def to_dict(self):
        return dict(self.iterable)

    def to_enum(self):
        return dict(enumerate(self.iterable))

    def head(self, num=3):
        lst = []
        for it in self.iterable:
            lst.append(it)
            if len(lst) > num:
                break
        return lst

    def quantile(self, p):
        """
        need self.iterable to be sortable
        """
        lst = sorted(self.iterable)
        return lst[int[len(lst) * p]]





from enum import Enum
import os


class Log(Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3


class Logger:
    def __init__(self, log_level=Log.INFO, file=None):
        self.log_level = Log(log_level)
        self.file = file

    def set_log_level(self, log_level):
        self.log_level = Log(log_level)

    def set_file(self, path):
        os.makedirs(path, exist_ok=True)
        self.file = open(path, 'a+')

    def print(self, msg, log_level=Log.INFO):
        if Log(log_level).value >= self.log_level.value:
            print(msg)
        if self.file is not None:
            print(msg, self.file)

    def debug(self, msg):
        self.print('[DEBUG] ' + msg, log_level=Log.DEBUG)

    def info(self, msg):
        self.print('[INFO] ' + msg, log_level=Log.INFO)

    def warn(self, msg):
        self.print('[WARNING]' + msg, log_level=Log.WARNING)

    def error(self, msg):
        self.print('[ERROR] ' + msg, log_level=Log.ERROR)



